﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading.Tasks;
using ArcGIS.Core.CIM;
using ArcGIS.Core.Data;
using ArcGIS.Core.Geometry;
using ArcGIS.Desktop.Editing;
using ArcGIS.Desktop.Mapping;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcgispro
{
    /// <summary>
    /// Class UpdateMeasurement manages the updates to the measurement layer
    /// </summary>
    class UpdateMeasurement
    {
        private MapObjectLayer layer;
        private PluginArcGIS viewerHost;
        private long updateTime;
        private long lastRequestTime;
        private bool isUpdating;
        private MeasureState measureState;
        private AList<MapObject> mapObjects;

        /// <summary>
        /// Create a new updater
        /// </summary>
        internal UpdateMeasurement(MapObjectLayer layer, PluginArcGIS viewerHost)
        {
            this.layer = layer;
            this.viewerHost = viewerHost;
            this.updateTime = 0;
            this.lastRequestTime = 0;
            this.isUpdating = false;
            this.measureState = null;
            this.mapObjects = null;
        }

        /// <summary>
        /// Called when the measurement has been changed and the display need to be updated
        /// </summary>
        internal void updatedMeasurement()
        {
            this.updateTime = DateTime.Now.Ticks;
        }

        /// <summary>
        /// Check for updates
        /// </summary>
        internal void checkForUpdates()
        {
            // already updating?
            if (this.isUpdating) return;
            // already up-to-date?
            if (this.lastRequestTime > this.updateTime) return;
            long currentTime = DateTime.Now.Ticks;
            // throttle the rate of update requests
            if ((this.lastRequestTime > 0) && (currentTime < this.lastRequestTime + 100 * TimeSpan.TicksPerMillisecond)) return;
            // throttle the rate of view changes (fast panning)
            if ((this.updateTime > 0) && (currentTime < this.updateTime + 100 * TimeSpan.TicksPerMillisecond)) return;
            // start the update task (to free the calling timer thread)
            this.isUpdating = true;
            this.lastRequestTime = currentTime;
            Task.Run(() => this.update());
        }

        /// <summary>
        /// Update
        /// </summary>
        private void update()
        {
            try
            {
                Message.WriteLine("Requesting measurement");
                this.measureState = this.viewerHost.getViewer().getMeasureState();
                if (this.measureState != null) this.mapObjects = this.measureState.mapObjects;
                if (this.mapObjects != null) Message.WriteLine("Got " + this.mapObjects.size() + " mapobjects");
                this.isUpdating = false;
                /* Update the layer */
                this.layer.update(this.mapObjects);
                this.viewerHost.RequestMapUpdate("updated measurement");
                return;
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Error while updating measurement: " + e);
                this.isUpdating = false;
                return;
            }
        }

        // Save (add) the current measurement to the active layer
        //
        internal void doSaveMeasurement(MapView mapView)
        {
            Message.WriteLine("Adding the measurement to the current layer");
            // we need to have a measurement
            if (this.measureState == null)
            {
                Message.WriteLine("There is no current measurement");
                return;
            }
            // print the attributes
            if (this.measureState.properties != null)
            {
                Message.WriteLine("The properties of the measurement are:");
                for (int i = 0; i < this.measureState.properties.size(); i++)
                {
                    MeasureProperty property = this.measureState.properties.get(i);
                    if (property.name != null) Message.WriteLine(">name: '" + property.name + "'");
                    if (property.label != null) Message.WriteLine(">label: '" + property.label + "'");
                    if (property.value != null) Message.WriteLine(">value: '" + property.value + "'");
                    if (property.unit != null) Message.WriteLine(">unit: '" + property.unit + "'");
                }
            }
            // get the current vector layer
            int layerCount = mapView.GetSelectedLayers().Count;
            if (layerCount != 1)
            {
                Message.WriteLine("There is no single selected layer to add to");
                return;
            }
            Layer currentLayer = mapView.GetSelectedLayers()[0];
            if (currentLayer is FeatureLayer == false)
            {
                Message.WriteLine("The current layer '" + currentLayer.Name + "' is not a feature layer");
                return;
            }
            FeatureLayer featureLayer = (FeatureLayer)currentLayer;
            // create the geometry for the layer
            esriGeometryType layerGeometryType = featureLayer.ShapeType;
            Geometry geometry = null;
            // get the (first) mapobject of the measurement
            if (this.measureState.mapObjects.size() < 1)
            {
                Message.WriteLine("The measure state does not have a mapobject");
                return;
            }
            MapObject mapObject = this.measureState.mapObjects.get(0);
            if (layerGeometryType == esriGeometryType.esriGeometryPoint)
            {
                if (mapObject.spatialPoint == null)
                {
                    Message.WriteLine("Cannot add a non-point measurement to a point layer");
                    return;
                }
                Message.WriteLine("Creating a point geometry");
                geometry = MapObjects.getPointGeometry(mapObject);
            }
            if (layerGeometryType == esriGeometryType.esriGeometryPolyline)
            {
                if (mapObject.spatialLine == null)
                {
                    Message.WriteLine("Cannot add a non-line measurement to a line layer");
                    return;
                }
                Message.WriteLine("Creating a line geometry");
                geometry = MapObjects.getLineGeometry(mapObject);
            }
            if (layerGeometryType == esriGeometryType.esriGeometryPolygon)
            {
                if (mapObject.spatialArea == null)
                {
                    Message.WriteLine("Cannot add a non-area measurement to an area layer");
                    return;
                }
                Message.WriteLine("Creating an area geometry");
                geometry = MapObjects.getAreaGeometry(mapObject);
            }
            if (geometry == null)
            {
                Message.WriteLine("The current layer geometry is neither point, line or polygon");
                return;
            }
            // convert the geometry into the CRS of the layer
            geometry = GeometryEngine.Instance.Project(geometry, featureLayer.GetSpatialReference());
            // make a new feature for the layer
            Dictionary<string, object> featureAttributes = new Dictionary<string, object>();
            featureAttributes["SHAPE"] = geometry;
            if (this.measureState.properties != null)
            {
                TableDefinition attributeFields = featureLayer.GetTable().GetDefinition();
                for (int i = 0; i < this.measureState.properties.size(); i++)
                {
                    MeasureProperty property = this.measureState.properties.get(i);
                    if ((property.name!=null) && (property.value!=null))
                    {
                        int attributeIndex = attributeFields.FindField(property.name);
                        if (attributeIndex >= 0)
                        {
                            Field attributeField = attributeFields.GetFields()[attributeIndex];
                            Message.WriteLine("Found attribute '" + attributeField.Name + "' type " + attributeField.FieldType + " for value '" + property.value + "'");
                            object attributeValue = null;
                            if (attributeField.FieldType == FieldType.String) attributeValue = (string)property.value;
                            else if (attributeField.FieldType == FieldType.Double) attributeValue = Convert.ToDouble(property.value);
                            else if (attributeField.FieldType == FieldType.Single) attributeValue = Convert.ToSingle(property.value);
                            else if (attributeField.FieldType == FieldType.Integer) attributeValue = (int)Convert.ToDouble(property.value);
                            else if (attributeField.FieldType == FieldType.SmallInteger) attributeValue = (int)Convert.ToDouble(property.value);
                            else Message.WriteLine("Attribute type not supported");
                            if (attributeValue!=null)
                            {
                                featureAttributes[attributeField.Name] = attributeValue;
                            }
                        }
                    }
                }
            }
            // add the features to the layer
            Message.WriteLine("Adding the new features to current layer '" + currentLayer.Name + "'");
            EditOperation editOperation = new EditOperation();
            editOperation.Name = "3D Mapping measurement save to feature";
            editOperation.Create(featureLayer, featureAttributes);
            bool success = editOperation.Execute();
            if (success) Message.WriteLine("The features were committed");
            else Message.WriteLine("The features were not committed");
        }
    }
}
